#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np

class MSeRegressor:
    def __init__(self, x=[], y=[]):
        self.x = np.array(x)          # data
        self.y = np.array(y)
        self.N = len(x)               # number of points
        
        self.w = []                   # augmented w, which includes w0, w1, ...

    # OLS
    def algebraicSol(self):
        pseudo = np.linalg.inv(np.matmul(self.x.T, self.x))
        self.w = np.matmul(np.matmul(pseudo, self.x.T), self.y)
        
    def J_n(self, n):
        return (2/self.N)*(np.matmul(self.w, self.x[n])-self.y[n])*self.x[n]
    
    # LMS
    def seqGDSol(self):
        while self.max_epoch:
            for n in range(self.N):
                self.w = self.w - self.lr*self.J_n(n)
            
            self.max_epoch -= 1
            if self.max_epoch % 5000 == 0 and self.max_epoch:
                print(f"Last {self.max_epoch} Epochs")
                
    # generally many function can be used as transformation: exp, sin, cos, 1/x, ...
    def kernelTransform(self, x):
        '''Only for one dimension data'''
        if self.kernel == 'polynomial':
            gen = np.hstack([np.reshape(x[:,1]**i, (-1, 1)) for i in range(2, self.d+1)])
            return np.hstack((x, gen))
        if self.kernel == 'sin':
            return np.sin(x)
        return x
        
    def fit(self, max_epoch=100, sol='algebra', kernel='linear', lr=0.01, d=3):
        self.sol = sol
        self.max_epoch = max_epoch
        self.kernel = kernel
        self.lr = lr
        self.d = d
        
        if self.kernel != 'linear':
            self.x = self.kernelTransform(self.x)
        
        self.w = np.zeros(len(self.x[0]))        # initialize w
        
        if self.sol == 'algebra':
            self.algebraicSol()
        elif self.sol == 'GD':
            self.seqGDSol()
        
        return [max_epoch, self.sol, self.lr]
    
    def predict(self, x):
        if self.kernel != 'linear':
            x = self.kernelTransform(x)
        result = np.matmul(x, self.w)
        return result


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
from .common import *
from .utils import *

class ANN:
    def __init__(self, x, y, model):
        self.x_val = np.array(x)
        self.y_val = np.array(y)
        self.model = model
        
        self.N = len(self.x_val)
        
        self.xdim = 1 if len(self.x_val.shape)==1 else len(self.x_val[0])
        self.ydim = 1 if len(self.y_val.shape)==1 else len(self.y_val[0])
    
    def initialize_w_b(self):
        self.W_val = []
        self.b_val = []
        pre = self.xdim
        for n in self.hidden_layer:
            self.W_val.append(np.random.randn(pre, n))
            self.b_val.append(np.random.randn(n))
            pre = n
    
    def create_network(self):
        self.x_node, self.y_node = Placeholder(), Placeholder()
        self.W_node = [Placeholder() for _ in range(len(self.hidden_layer))]
        self.b_node = [Placeholder() for _ in range(len(self.hidden_layer))]

        self.linear = []
        self.activation = []
        pre = self.x_node
        for i in range(len(self.hidden_layer)):
            self.linear.append(Linear(pre, self.W_node[i], self.b_node[i]))
            if i == len(self.hidden_layer)-1 and self.model == 'regressor':
                break
            if self.activation_func == 'Sigmoid':
                self.activation.append(Sigmoid(self.linear[-1]))
            elif self.activation_func == 'Relu':
                self.activation.append(Relu(self.linear[-1]))
            elif self.activation_func == 'Tanh':
                self.activation.append(Tanh(self.linear[-1]))
            pre = self.activation[-1]
        self.cost = MSE(self.y_node, self.activation[-1] if self.model=='classifier' else self.linear[-1])
    
    def create_traning_graph(self):
        self.feed_dict = {
            self.x_node: self.x_val,
            self.y_node: self.y_val
        }
        for i in range(len(self.hidden_layer)):
            self.feed_dict[self.W_node[i]] = self.W_val[i]
            self.feed_dict[self.b_node[i]] = self.b_val[i]

        graph = feed_dict_2_graph(self.feed_dict)    # network graph
        self.sorted_graph = topology(graph)          # sorted graph
        self.trainables = self.W_node+self.b_node    # trainable parameters
        
    def fit(self, hidden_layer=[1], activation='Sigmoid', optimizer='SGD', lr=1e-1, epochs=1000, batch_size=16, verbose=True):
        self.hidden_layer = hidden_layer+[self.ydim]
        self.activation_func = activation
        self.optimizer = optimizer
        
        # step 1: initialize weights and bias
        self.initialize_w_b()
        # step 2: Neural network nodes
        self.create_network()
        # step 3: create graph and trainable nodes
        self.create_traning_graph()
        
        # step 4: train
        steps_per_epoch = len(self.x_val) // batch_size
        losses = []

        # for Adam
        it = 0
        
        for i in range(1, epochs+1):
            loss = 0
            index = list(range(len(self.x_val)))
            np.random.shuffle(index)
            self.x_val = self.x_val[index]
            self.y_val = self.y_val[index]
            for j in range(steps_per_epoch):
                it += 1
                # Step 4.1: Randomly sample a batch of examples and Reset value
                # index = np.random.choice(self.N, batch_size)
                self.x_node.value = self.x_val[j*batch_size:(j+1)*batch_size]
                self.y_node.value = self.y_val[j*batch_size:(j+1)*batch_size]

                # Step 4.2: forward
                for n in self.sorted_graph:
                    n.forward()

                # Step 4.3: backward
                for n in self.sorted_graph[::-1]:
                    n.backward()

                # Step 4.4: optimization
                for t in self.trainables:
                    t.optimize(optimizer=optimizer, lr=lr, it=it)

                # Step 5: update current loss
                loss += self.sorted_graph[-1].value

            if i % 1 == 0: 
                if verbose:
                    print("Epoch: {}, Loss: {:.4f}".format(i, loss/steps_per_epoch))
                losses.append(loss/steps_per_epoch)
        
        return [hidden_layer, activation, optimizer, lr, epochs, batch_size, losses]
        
    def predict(self, inputs):
        self.x_node.value = inputs
        for n in self.sorted_graph[:-1]:
            n.forward()
        if self.model == 'classifier':
            return np.argmax(self.sorted_graph[-2].value, axis=1)
        return self.sorted_graph[-2].value



#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np
import matplotlib.pyplot as plt

class Plot:
    # for complex kernel
#     def plot_kernel(self, x):
#         if self.kernel == 'RBF': # RBF kernel's Φ(x) is infinite dimensional
# #             return np.exp(-self.gamma*np.sum((self.u-x)*(self.u-x), axis=1))
#             # avoid for loop to accelerate computation
#             return np.exp(-self.gamma*((np.sum(self.u*self.u, axis=1) - 2*np.matmul(self.u, x.T).T).T + np.sum(x*x, axis=1)))
#         if self.kernel == 'quadratic':
#             return (np.matmul(self.u, x.T)+1)*(np.matmul(self.u, x.T)+1)
#         return np.matmul(self.u, x.T)
    def K_(self, x):
        return np.matmul(self.u, x.T)
    
    # prediction
    def classify(self, x):
#         result = np.sum(np.multiply(self.lambda_*self.z, np.array([self.plot_kernel(i, x) for i in range(self.N)]).T).T, axis=0) + self.w0
        result = np.sum((self.lambda_*self.z*self.K_(x).T).T, axis=0) + self.w0
        result[result>0], result[result<0]= 0, 1 # class 1 & class 2
        return result
    
    # helper function for plot svm on 2 dim
    def plot(self):
        if len(self.u[0]) > 2:
            print("Cannot draw high dimension!")
            return
        
#         if len(self.w) > 2:
#             self.i0 = np.argmax(self.w)
#             self.i1 = np.argmax(np.delete(self.w, self.i0))
#         plt.clf() # clear previous graph

        class_names = ['Class 1', 'Class 2']

        # Set the feature range for plotting
        max_x1 = np.ceil(np.max(self.u[:, 0])) + 1.0
        min_x1 = np.floor(np.min(self.u[:, 0])) - 1.0
        max_x2 = np.ceil(np.max(self.u[:, 1])) + 1.0
        min_x2 = np.floor(np.min(self.u[:, 1])) - 1.0

        xrange = (min_x1, max_x1)
        yrange = (min_x2, max_x2)

        # step size for how finely you want to visualize the decision boundary.
        inc = int(max_x1 - min_x1)/200

        # generate grid coordinates. This will be the basis of the decision
        # boundary visualization.
        (x1, x2) = np.meshgrid(np.arange(xrange[0], xrange[1] + inc / 100, inc),
                               np.arange(yrange[0], yrange[1] + inc / 100, inc))

        # size of the (x1, x2) image, which will also be the size of the
        # decision boundary image that is used as the plot background.
        image_size = x1.shape
        # make (x1, x2) pairs as a bunch of row vectors.
        grid_2d = np.hstack((x1.reshape(x1.shape[0] * x1.shape[1], 1, order='F'),
                             x2.reshape(x2.shape[0] * x2.shape[1], 1, order='F')))

        # Labels for each (x1, x2) pair.
        pred_label = self.classify(grid_2d)
        pred_label = np.array(pred_label)

        # reshape the idx (which contains the class label) into an image.
        decision_map = pred_label.reshape(image_size, order='F')

        if not self.ani:
            self.fig = plt.figure(figsize=(6, 6)) # create new
        else:
            if not hasattr(self, 'fig'):
                self.fig = plt.figure(figsize=(6, 6)) # create new
        
        # show the image, give each coordinate a color according to its class
        # label
        im = plt.imshow(decision_map, 
                        vmin=0, 
                        vmax=9,
                        cmap='Pastel1', 
                        extent=[xrange[0], xrange[1], yrange[0], yrange[1]], 
                        origin='lower').findobj()

        # plot the class training data.
        data_point_styles = ['rx', 'bo', 'g*']
        for i in range(2):
            im += plt.plot(self.u[self.z == (1 if not i else -1), 0],
                    self.u[self.z == (1 if not i else -1), 1],
                    data_point_styles[i],
                    label=class_names[i])
        if not self.ani:
            plt.legend()
        
#         # plot support vectors
#         lmd = self.lambda_[self.lambda_>0]
#         for i, (x, y) in enumerate(self.u[self.lambda_>0][:,[0,1]]):
#             if round(lmd[i], 5) > 0:
#                 im += plt.scatter(x, y, s=300, alpha = .5).findobj()

        # only for 2 dim draw
        if self.kernel == 'linear':
            # draw arrow of S1
            bbox_props = dict(boxstyle="rarrow", fc=(0.8, 0.9, 0.9), ec="r", lw=2)
            degree = np.arctan(self.w[1] / self.w[0])*180/np.pi
            t = plt.text((max_x1*0.3+min_x1*0.7), 
                              (self.w[0]*((max_x1*0.3+min_x1*0.7)) + self.w0)/(-self.w[1]), 
                              "S2" if self.w[0] < 0 else "S1", rotation=degree, size=15, bbox=bbox_props)
            bb = t.get_bbox_patch()
            bb.set_boxstyle("rarrow", pad=0.2)
        
            # draw bound of distance
            b = self.b[0] if type(self.b) != type(1) else self.b # self.b was changed in algebraic method
            # bound 1
            x1 = [min_x1, max_x1]
            y1 = [(self.w[0]*x + self.w0 - b)/(-self.w[1]) for x in x1]
            for i, y in enumerate(y1):
                if y > max_x2: # check up bound
                    x1[i], y1[i] = (b - self.w[1]*max_x2 - self.w0)/self.w[0], max_x2
                if y < min_x2: # check bottom bound
                    x1[i], y1[i] = (b - self.w[1]*min_x2 - self.w0)/self.w[0], min_x2
            plt.plot(x1, y1, 'r.:', label='bound S1', linewidth=1, alpha=0.8)
            # bound 2
            x2 = [min_x1, max_x1]
            y2 = [(self.w[0]*x + self.w0 + b)/(-self.w[1]) for x in x2]
            for i, y in enumerate(y2):
                if y > max_x2: # check up bound
                    x2[i], y2[i] = (-b - self.w[1]*max_x2 - self.w0)/self.w[0], max_x2
                if y < min_x2: # check bottom bound
                    x2[i], y2[i] = (-b - self.w[1]*min_x2 - self.w0)/self.w[0], min_x2
            plt.plot(x2, y2, 'b.:', label='bound S2', linewidth=1, alpha=0.8)

        plt.tight_layout()
        self.ims.append(im)

        return self.fig


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import random
import numpy as np

class SMoSVR:
    def __init__(self, x=[], y=[]):
        self.x = np.array(x)          # data
        self.y = np.array(y)
        self.N = len(x)               # number of points
        
        self.beta = []

        self.w = []                   # optimal w
        self.w0 = None                # optimal bias term w0

    # n should be not equal to m
    def random_n(self, m):
        n = random.randint(0, self.N-1)
        while n == m:
            n = random.randint(0, self.N-1)
        return n

    def K_(self, x):
        if self.kernel == 'RBF':
            return np.exp(-self.gamma*((np.sum(self.x*self.x, axis=1) - 2*np.matmul(self.x, x.T).T).T + np.sum(x*x, axis=len(x.shape)-1)))
        if self.kernel == 'polynomial':
            return (self.gamma*np.matmul(self.x, x.T)+self.r)**self.d
        return np.matmul(self.x, x.T)
    
    def K_ij(self, i, j):
        if self.kernel == 'RBF': # RBF kernel's Φ(x) is infinite dimensional
            return np.exp(-self.gamma*np.sum((self.x[i]-self.x[j])*(self.x[i]-self.x[j])))
        if self.kernel == 'polynomial':
            return (self.gamma*np.dot(self.x[i], self.x[j])+self.r)**self.d
        return np.dot(self.x[i], self.x[j])
    
    def f_k(self, k):
        return np.sum((self.beta*self.K_(self.x[k]).T).T, axis=0) + self.w0
    
    def sign(self, x):
        return x/abs(x) if x else 0
    
    def update_w0(self, idx, m, n, o_m, n_m, o_n, n_n):
        return self.w0 + self.y[idx] - self.f_k(idx) + (o_m-n_m)*self.K_ij(m,idx) + (o_n-n_n)*self.K_ij(idx,n)

    def fit(self, max_epoch=100, kernel='linear', C=1.0, epsilon=0.1, gamma=1, d=2, r=1):
        self.C = C
        self.kernel = kernel
        self.gamma = gamma
        self.d = d
        self.r = r

        self.beta = np.zeros((self.N))         # initialize beta
        self.w0 = 0                            # initialize w0
        while max_epoch:
            prev_bata = np.copy(self.beta)
            for m in range(self.N):
                # step 1: randomly choose n != m
                n = self.random_n(m)
                
                old_n, old_m = self.beta[n], self.beta[m]
                
                # step 2: calculate temporary beta_n
                alpha = old_n + old_m
                eta = self.K_ij(m,m) + self.K_ij(n,n) - 2*self.K_ij(m,n)
                if not eta: # in case there are two points have exactly the same coordinates
                    continue
                cort = 2*epsilon/eta
                temp_n = old_n + (self.y[n]-self.y[m]+self.f_k(m)-self.f_k(n))/eta
                temp_m = alpha - temp_n
                
                # step 3: apply correction to temporary beta_n
                if temp_m*temp_n < 0:
                    if abs(temp_m) >= cort and abs(temp_n) >= cort:
                        temp_n += self.sign(temp_m)*cort
                    else:
                        temp_n = alpha if abs(temp_n) > abs(temp_m) else 0
                
                # step 4: crop temporary beta_n
                L = max(alpha-self.C, -self.C)
                H = min(self.C, alpha+self.C)
                beta_n = min(max(temp_n, L), H)
                
                # step 5: calculate new beta_m
                beta_m = alpha - beta_n
                
                # step 6: calculate new w0 (slack form)
                if not beta_m:
                    w0 = self.update_w0(m, m, n, old_m, beta_m, old_n, beta_n)
                elif not beta_n:
                    w0 = self.update_w0(n, m, n, old_m, beta_m, old_n, beta_n)
                else:
                    w0_1 = self.update_w0(m, m, n, old_m, beta_m, old_n, beta_n)
                    w0_2 = self.update_w0(n, m, n, old_m, beta_m, old_n, beta_n)
                    w0 = (w0_1+w0_2)/2
                
                # update all together
                self.beta[m] = beta_m
                self.beta[n] = beta_n
                self.w0 = w0
            
            max_epoch -= 1
            if max_epoch % 100 == 0 and max_epoch:
                print(f"Last {max_epoch} Epochs")
        return [max_epoch, C, epsilon]
    
    def predict(self, x):
        result = np.sum((self.beta*self.K_(x).T).T, axis=0) + self.w0
        return result


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np
from .common import Placeholder

# use topology to generate sorted graph for better forward and backward operations
def topology(graph):
    sorted_node = []
    
    while len(graph) > 0: 

        all_inputs = []
        all_outputs = []
        
        for n in graph:
            all_outputs += graph[n]
            all_inputs.append(n)
        
        all_outputs = set(all_outputs)
        all_inputs = set(all_inputs)
    
        # find out node without indegree
        need_remove = all_inputs - all_outputs
    
        if len(need_remove) > 0: 
            # only choose one
            node = np.random.choice(list(need_remove))

            need_to_visited = [node]

            # when this is the last input node, add its all output to make sure all nodes are in sorted_node
            if len(graph) == 1: need_to_visited += graph[node]
            
            # delete the chosen node
            graph.pop(node)
            
            sorted_node += need_to_visited
        else: # is cycle
            break
        
    return sorted_node

# generate graph from dict structure
def feed_dict_2_graph(feed_dict):
    computing_graph = dict()
    
    nodes = [n for n in feed_dict]
    
    while nodes:
        n = nodes.pop(0) 
        
        if isinstance(n, Placeholder):
            n.value = feed_dict[n]
        
        if n in computing_graph:
            continue
        computing_graph[n] = list()
        
        for m in n.outputs:
            computing_graph[n].append(m)
            nodes.append(m)
    
    return computing_graph


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
from .Plot import *

class AlgebraicSVC(Plot):
    def __init__(self, z=[], u=[], b=1):
        self.z = np.array(z)  # classes, only for two: 1 or -1
        self.u = np.array(u)  # data
        self.N = len(z)       # number of points
        
        # parameters for calc lambda
        self.A = []           # A in Aρ=b
        self.b = b            # margin, which is also one value for b in Aρ=b
        self.lambda_ = []     # lambda
        self.miu = None       # miu
        
        # parameters for calc w
        self.w = []           # optimal w
        self.w0 = None        # optimal bias term w0
        
        # fixed, do not change
        self.ani = False
        self.ims = []
        self.kernel = 'linear'
        
    # calculate A and b in Aρ=b
    def cal_A_b(self):
        if type(self.z) == list and type(self.u) == list:
            return None, None
        if type(self.A) != list and type(self.b) != list:
            return self.A, self.b
        # calculate A
        A = []
        for i in range(self.N):
            Ai = []
            for j in range(self.N):
                term = self.z[j]*self.z[i]*np.matmul(self.u[j], self.u[i])
                Ai.append(term)
            Ai.append(-self.z[i])
            A.append(Ai)
        A.append([self.z[i] for i in range(self.N)] + [0])
        self.A = np.array(A)
        # calculate b
        self.b = np.full(self.N+1, self.b)
        self.b[-1] = 0 # this is for sum of z_i*lambda_i = 0
        return self.A, self.b
    
    # Use NumPy to invert matrix, and to calculate lambda and µ
    def calc_lambda_miu(self):
        if self.miu != None:
            return self.lambda_, self.miu
        A, b = self.cal_A_b()
        if type(A) == type(None) or type(b) == type(None):
            return None, None
        # calculate ρ in Aρ=b
        ρ = np.matmul(np.linalg.inv(A), b)
        self.lambda_, self.miu = ρ[:-1], ρ[-1]
        return self.lambda_, self.miu
    
    # assign certain lambda to zero
    def calc_lambda_miu_with_zero_lambda(self, zeros=[]):
        # after deleting, A becomes NxN, b becomes Nx1
        A, b = np.delete(self.A, zeros, axis=1), np.delete(self.b, zeros) # delete the columns in list of zeros
        A = np.delete(A, zeros, axis=0) # delete the rows in list of zeros
        if type(A) == type(None) or type(b) == type(None):
            return None, None
        # calculate ρ in Aρ=b
        ρ = np.matmul(np.linalg.inv(A), b)
        self.lambda_, self.miu = ρ[:-1], ρ[-1]
        zeros.sort()
        for loc in zeros:
            self.lambda_ = np.insert(self.lambda_, loc, 0) # insert the fixed zero lambda
        return self.lambda_, self.miu
    
    # Calculate the optimal (nonaugmented) weight vector w∗ and w0
    def calc_w_w0(self):
        # calculate for optimal w
        self.w = np.full(2, 0.0)
        for i in range(self.N):
            self.w += self.lambda_[i]*self.z[i]*self.u[i]
        # calculate for optimal w0
        for i in range(self.N):
            if self.lambda_[i] != 0:
                self.w0 = self.b[0]/self.z[i] - np.matmul(self.w, self.u[i])
                break
        return self.w, self.w0
    
    # Check lambda satisfies the KKT conditions
    def check_KKT_lambda(self):
        sum_lambda_z = 0
        for i in range(self.N):
            sum_lambda_z += self.lambda_[i]*self.z[i]
            if self.lambda_[i] < 0:
                print(f"[Check Error] Lambda {self.lambda_[i]} is less than zero!!!")
                return False
        if abs(sum_lambda_z) > 1e-10:
            print(f"[Check Error] Sum of product of lambda and z ({sum_lambda_z}) is not zero!!!")
            return False
        return True

    # Check that the resulting w and w0 satisfy the KKT conditions
    def check_KKT_w_w0(self):
        sum_lambda_z_u = np.full(2, 0.0)
        for i in range(self.N):
            check1 = self.lambda_[i]*(self.z[i]*(np.matmul(self.w, self.u[i]) + self.w0)-self.b[0])
            if abs(check1) > 1e-10:
                print(f"[Check Error] λi*[Zi(w*ui + w0)-1] != 0 ({check1})!!!")
                return False
            check2 = self.z[i]*(np.matmul(self.w, self.u[i]) + self.w0) - self.b[0]
            check2 = 0 if abs(check2) < 1e-10 else check2
            if check2 < 0:
                print(f"[Check Error] Zi(w*ui + w0)-1 < 0 ({check2})!!!")
                return False
            sum_lambda_z_u += self.lambda_[i]*self.z[i]*self.u[i]
        if sum(sum_lambda_z_u != self.w):
            print(f"[Check Error] W* != λi*Zi*ui {sum_lambda_z_u} != {self.w}!!!")
            return False
        return True


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np

class PCA:
    def __init__(self, x):
        self.x = np.array(x)
        self.dim = 1 if len(x.shape)==1 else len(x[0])
    
    def fit(self, n_components=1):
        self.n_components = n_components
        # step 1: calculate mean
        m = np.mean(self.x, axis=0)
        # step 2: calculate sample covariance matrix
        cov = np.matmul((self.x - m).T, (self.x - m)) / len(self.x)
        # step 3: calculate eigens
        self.eigenvalues, self.normalized_eigenvectors = np.linalg.eig(cov)
        # step 4: find maximum eigenvalues
        max_index = np.argsort(self.eigenvalues)[::-1]
        keep = min(self.n_components, self.dim)
        self.index = max_index[:keep]
    
    def transform(self, inputs):
        return np.matmul(inputs, self.normalized_eigenvectors[:,self.index])


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np

class KNearestNeighborsClassifier:
    def __init__(self, data, labels):
        self.data = np.array(data)
        self.labels = np.array(labels)       # starts from 0
    
    def fit(self, k):
        self.k = k

    def self_cdist(self, inputs):
        dist = []
        dim = len(self.data[0])
        calc = np.hstack([inputs for _ in range(len(self.data))]) - self.data.reshape(1, -1)
        for i in range(len(self.data)):
            class_ = calc[:,i*dim:(i+1)*dim]
            dist.append(np.linalg.norm(class_, axis=1).reshape(-1,1))
        return np.hstack(dist)
    
    def getMode(self, x):
        vals, counts = np.unique(x, return_counts=True)
        index = np.argmax(counts)
        return vals[index]
    
    def predict(self, inputs):
        # step 1: calculate the dist between each point and each data in dataset
        dist_mat = self.self_cdist(inputs)
        
        # step 2: for each point, pick the nearest k data in dataset
        index = np.argsort(dist_mat, axis=1)[:,:self.k]
        
        # step 3: for each point, calculate assigned class by nearest k data's classes
        preds = list(map(self.getMode, self.labels.reshape(-1,1)[index].reshape(-1,self.k)))
        return np.array(preds)


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
from .SMoSVC import *

class MultiSVC:
    def __init__(self, z=[], u=[]):
        self.z = z            # classes, 0, 1, 2...
        self.u = u            # data
        self.N = len(z)       # number of points
        
        self.z = np.array(self.z)
        self.u = np.array(self.u)
        
        self.c = len(set(self.z))
        self.svms = []

        self.kernel = ''

        for c in range(self.c):
            cur_z = np.copy(self.z)
            cur_z[cur_z!=c] = -1
            cur_z[cur_z==c] = 1
            cur_u = np.copy(self.u)
            self.svms.append(SMoSVC(z=cur_z, u=cur_u, b=1))
        
    def fit(self, max_epoch=10, kernel='linear', gamma=0.01):
        """
        Args:
            max_epoch: Maximum training epoch if linear KKT cannot be satisfied.
            kernel: Kernel function. Default is linear.
            C: Upper Limit of lambda. A parameter of soft margin/slack-SVM.
            epsilon: Threshold of lambda value change, if which is less than this & KKT is satisfied, then traning ends in advance.
            ani: Launch traning process record as .gif file. This may cause much slower traning because of drawing in each epoch.
            gamma: Parameter of RBF kernel.
            d: Parameter of polynomial kernel.

        Returns:
            Return accuracy on traning set
        """
        self.kernel = kernel
        for c in range(self.c):
            self.svms[c].fit(max_epoch=max_epoch, kernel=self.kernel, gamma=gamma)

        # pred_label = self.predict(self.u)
        return [max_epoch, kernel, gamma] # [round(np.sum(pred_label == self.z)/self.N, 5)]

    def predict(self, x):
        results = []
        for c in range(self.c):
            results.append(np.sum((self.svms[c].lambda_*self.svms[c].z*self.svms[c].K_(x).T).T, axis=0) + self.svms[c].w0)
        results = np.array(results)
        pred_label = np.argmax(results, axis=0)
        return pred_label
    
    def plot(self):
        # Set the feature range for plotting
        max_x1 = np.ceil(np.max(self.u[:, 0])) + 1.0
        min_x1 = np.floor(np.min(self.u[:, 0])) - 1.0
        max_x2 = np.ceil(np.max(self.u[:, 1])) + 1.0
        min_x2 = np.floor(np.min(self.u[:, 1])) - 1.0

        xrange = (min_x1, max_x1)
        yrange = (min_x2, max_x2)

        # step size for how finely you want to visualize the decision boundary.
        inc = int(max_x1 - min_x1)/200

        # generate grid coordinates. This will be the basis of the decision
        # boundary visualization.
        (x1, x2) = np.meshgrid(np.arange(xrange[0], xrange[1] + inc / 100, inc),
                               np.arange(yrange[0], yrange[1] + inc / 100, inc))

        # size of the (x1, x2) image, which will also be the size of the
        # decision boundary image that is used as the plot background.
        image_size = x1.shape
        # make (x1, x2) pairs as a bunch of row vectors.
        grid_2d = np.hstack((x1.reshape(x1.shape[0] * x1.shape[1], 1, order='F'),
                             x2.reshape(x2.shape[0] * x2.shape[1], 1, order='F')))

        pred_label = self.predict(grid_2d)

        # reshape the idx (which contains the class label) into an image.
        decision_map = pred_label.reshape(image_size, order='F')

        self.fig = plt.figure(figsize=(6, 6)) # create new
        
        # show the image, give each coordinate a color according to its class
        # label
        plt.imshow(decision_map, 
                        vmin=0, 
                        vmax=9,
                        cmap='Pastel1', 
                        extent=[xrange[0], xrange[1], yrange[0], yrange[1]], 
                        origin='lower')

        # plot the class training data.
        data_point_styles = ['rx', 'bo', 'g*']
        for i in range(self.c):
            plt.plot(self.u[self.z == i, 0],
                    self.u[self.z == i, 1],
                    data_point_styles[i],
                    label=f'Class {i}')
        plt.legend()
        plt.tight_layout()

        return self.fig


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import random
import numpy as np

class Perceptron:
    def __init__(self, data, labels):
        self.data = np.array(data)     # augmented data
        self.labels = np.array(labels) # starts from 0
        
        self.w = []                    # augmented notation

    # Initialize indicator function
    def indicator(self, x):
        if type(x) != bool and type(x) != np.bool_:
            return x*np.full(len(x), 1)
        return 1 if x == True else 0
    
    # criterion function
    def J(self, trains, labels, ind=0):
        if ind == 1:
            print(self.w)
        z = np.full(len(trains), 1)
        z[labels != 0] = -1
        com = np.matmul(trains, self.w)*z
        return -np.sum(self.indicator(com<=0)*com)
    
    # Derivative of sequential criterion function
    def J_i_derivative(self, x, label):
        z = 1 if label == 0 else -1
        return -self.indicator(np.sum(self.w*z*x)<=0)*z*x
    
    def J_batch_derivative(self, data, labels):
        z = np.copy(labels)
        z[labels==0] = 1
        z[labels!=0] = -1
        return -np.sum((self.indicator(np.matmul(data, self.w)*z<=0)*z).reshape(-1,1)*data, axis=0)
    
    # theoretically, exp, sin can also be used, but they perform wrose when it comes to classification
    def nonlinearTransform(self, x):
        if self.kernel == 'quadratic':
            return np.array(list(map(lambda x: [1, x[1], x[2], x[1]*x[2], x[1]*x[1], x[2]*x[2]], x)))
        return x
    
    # SequentialGD
    def fit(self, iters = 10000, kernel='linear', GD='SGD'):
        n = len(self.data) # total num of data
        self.kernel = kernel
        
        if self.kernel != 'linear':
            self.data = self.nonlinearTransform(self.data)
        
        # step 1: Shuffle dataset and corresponding labels
        indices = list(range(n))
        random.shuffle(indices)
        trains = self.data[indices]
        labels = self.labels[indices]
        
        # step 2: Initialize w, lr
        self.w = np.full(len(self.data[0]), 0.1)
        lr = 0.0001
        
        smallest_J = float("inf")
        best_W = np.copy(self.w)

        start = 0
        size = 5
        losses = []
        loss_light = []
        # step 3: sequential gradient descent
        while (iters):
            if GD == 'SGD':
                self.w -= lr*self.J_i_derivative(trains[iters%n], labels[iters%n])
            elif GD == 'MBGD': # mini batch
                if start + size < n:
                    end = start + size
                else:
                    start = 0
                    end = start + size
                self.w -= lr*self.J_batch_derivative(trains[start:end], labels[start:end])
            elif GD == 'BGD': # batch GD
                self.w -= lr*self.J_batch_derivative(trains, labels)
            cur_J = self.J(trains, labels)
            losses.append(cur_J)
            if not cur_J:
                print("Data is linearly separable")
                loss_light.append(1)
                break
            iters -= 1
            if (cur_J < smallest_J):
                smallest_J = cur_J
                best_W = np.copy(self.w)
                loss_light.append(1)
            else:
                loss_light.append(0)
        
        self.w = best_W
        return [iters, losses, loss_light]
        
    def predict(self, x):
        x = np.insert(x, 0, [1 for _ in range(len(x))], axis=1) # augmented notation
        if self.kernel != 'linear':
            x = self.nonlinearTransform(x)
        
        pred = np.matmul(x, self.w)
        pred[pred>0] = 0 # class 1
        pred[pred<0] = 1 # class 2
        return pred


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np

class KNearestNeighborsRegressor:
    def __init__(self, x, y):
        self.x = np.array(x)
        self.y = np.array(y)
    
    def fit(self, k):
        self.k = k

    def self_cdist(self, inputs):
        dist = []
        dim = len(self.x[0])
        calc = np.hstack([inputs for _ in range(len(self.x))]) - self.x.reshape(1, -1)
        for i in range(len(self.x)):
            class_ = calc[:,i*dim:(i+1)*dim]
            dist.append(np.linalg.norm(class_, axis=1).reshape(-1,1))
        return np.hstack(dist)
    
    def predict(self, inputs, weights='uniform'):        
        # step 1: calculate the dist between each point and each data in dataset
        dist_mat = self.self_cdist(inputs)
        
        # step 2: for each point, pick the nearest k data in dataset
        index = np.argsort(dist_mat, axis=1)[:,:self.k]

        # step 3: for each point, calculate y vals by nearest k data's y's
        vals = self.y.reshape(-1,1)[index].reshape(-1,self.k)
        if weights == 'uniform':
            preds = np.mean(vals, axis=1)
        elif weights == 'distance':
            ws = 1/(np.array(list(map(sorted, dist_mat)))[:,:self.k]+0.1) # weights
            preds = np.zeros(vals[:,0].shape).reshape(-1,1)
            for i in range(self.k):
                preds += (vals[:,i]*ws[:,i]).reshape(-1,1)
            preds /= np.sum(ws, axis=1).reshape(-1,1)
        else:
            return None
        return preds


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np

class NearestMean:
    def __init__(self, data, labels):
        self.data = np.array(data)
        self.labels = np.array(labels)       # starts from 0
        
        self.c = len(np.unique(self.labels)) # num of classes
        self.sample_means = []
    
    def fit(self):
        self.sample_means = []
        for i in range(self.c):
            self.sample_means.append(np.mean(self.data[self.labels==i], axis=0))
        self.sample_means = np.array(self.sample_means)

    def predict(self, inputs):
        pred = []
        for i, xy in enumerate(inputs):
            my_eu = []
            for j in range(self.c):
                # or np.linalg.norm(xy - self.sample_means[j])
                my_eu.append(np.sqrt(np.sum(np.square(xy - self.sample_means[j]))))
            pred.append(np.argmin(my_eu))
        return np.array(pred)


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm

def self_cdist(nclass, xy, sample_mean):
    dim = len(xy[0])
    dis = []
    for i in range(nclass):
        class_ = (np.hstack([xy for _ in range(nclass)]) - sample_mean.reshape(1, -1))[:,i*dim:(i+1)*dim]
        dis.append(np.linalg.norm(class_, axis=1).reshape(-1,1))
    return np.hstack(dis)

def plotDecBoundByMean(training, label_train, sample_mean, title=""):
    nclass =  len(np.unique(label_train))

    # Set the feature range for ploting
    max_x = np.ceil(max(training[:, 0])) + 1
    min_x = np.floor(min(training[:, 0])) - 1
    max_y = np.ceil(max(training[:, 1])) + 1
    min_y = np.floor(min(training[:, 1])) - 1

    xrange = (min_x, max_x)
    yrange = (min_y, max_y)

    inc = 0.005

    (x, y) = np.meshgrid(np.arange(xrange[0], xrange[1]+inc/100, inc), 
                         np.arange(yrange[0], yrange[1]+inc/100, inc))

    # size of the (x, y) image, which will also be the size of the
    # decision boundary image that is used as the plot background.
    image_size = x.shape
    xy = np.hstack((x.reshape(x.shape[0]*x.shape[1], 1, order='F'), 
                    y.reshape(y.shape[0]*y.shape[1], 1, order='F'))) # make (x,y) pairs as a bunch of row vectors.

    # alternative: dist_mat = scipy.spatial.distance.cdist(xy, sample_mean)
    dist_mat = self_cdist(nclass, xy, sample_mean)
    
    pred_label = np.argmin(dist_mat, axis=1)

    decisionmap = pred_label.reshape(image_size, order='F')

    plt.imshow(decisionmap, extent=[xrange[0], xrange[1], yrange[0], yrange[1]], origin='lower')

    # plot the class training data.
    plt.plot(training[label_train == 0, 0],training[label_train == 0, 1], 'rx')
    plt.plot(training[label_train == 1, 0],training[label_train == 1, 1], 'go')
    if nclass == 3:
        plt.plot(training[label_train == 2, 0],training[label_train == 2, 1], 'b*')

    # include legend for training data
    if nclass == 3:
        l = plt.legend(('Class 1', 'Class 2', 'Class 3'), loc=2)
    else:
        l = plt.legend(('Class 1', 'Class 2'), loc=2)
    plt.gca().add_artist(l)

    # plot the class mean vector.
    m1, = plt.plot(sample_mean[0,0], sample_mean[0,1], 'rd', markersize=12, markerfacecolor='r', markeredgecolor='w')
    m2, = plt.plot(sample_mean[1,0], sample_mean[1,1], 'gd', markersize=12, markerfacecolor='g', markeredgecolor='w')
    if nclass == 3:
        m3, = plt.plot(sample_mean[2,0], sample_mean[2,1], 'bd', markersize=12, markerfacecolor='b', markeredgecolor='w')

    # include legend for class mean vector
    if nclass == 3:
        l1 = plt.legend([m1,m2,m3],['Class 1 Mean', 'Class 2 Mean', 'Class 3 Mean'], loc=4)
    else:
        l1 = plt.legend([m1,m2], ['Class 1 Mean', 'Class 2 Mean'], loc=4)

    plt.gca().add_artist(l1)
    plt.title(title)
    plt.show()

def plotKMean(training, k, sample_mean, title=""):
    plt.figure(figsize=(8,8))

    # Set the feature range for ploting
    max_x = np.ceil(max(training[:, 0])) + 1
    min_x = np.floor(min(training[:, 0])) - 1
    max_y = np.ceil(max(training[:, 1])) + 1
    min_y = np.floor(min(training[:, 1])) - 1

    xrange = (min_x, max_x)
    yrange = (min_y, max_y)

    inc = 0.005

    (x, y) = np.meshgrid(np.arange(xrange[0], xrange[1]+inc/100, inc), np.arange(yrange[0], yrange[1]+inc/100, inc))

    # size of the (x, y) image, which will also be the size of the
    # decision boundary image that is used as the plot background.
    image_size = x.shape
    xy = np.hstack( (x.reshape(x.shape[0]*x.shape[1], 1, order='F'), y.reshape(y.shape[0]*y.shape[1], 1, order='F')) ) # make (x,y) pairs as a bunch of row vectors.

    # alternative: dist_mat = scipy.spatial.distance.cdist(xy, sample_mean)
    dist_mat = self_cdist(k, xy, sample_mean)
    
    pred_label = np.argmin(dist_mat, axis=1)

    decisionmap = pred_label.reshape(image_size, order='F')

    plt.imshow(decisionmap, extent=[xrange[0], xrange[1], yrange[0], yrange[1]], origin='lower', cmap=plt.cm.Paired)

    # plot the class training data.
    plt.plot(training[:, 0], training[:, 1], 'g.', markersize=2)

    # plot the class mean vector.
    plt.plot(sample_mean[:,0], sample_mean[:,1], 'x', markersize=24, color='w')

    plt.title(title)
    plt.show()


def plotBoundByDecisionFunction(training, label_train, decision_function, inc=0.1, augmented=False):
    nclass = len(np.unique(label_train))

    x1_idx, x2_idx = 0, 1
    # for criterion function: g(x) = w*x, which w and x are augmented, 1 in x is at 0th place
    if augmented:
        x1_idx, x2_idx = 1, 2

    # Set the feature range for plotting
    max_x1 = np.ceil(np.max(training[:, x1_idx])) + 1.0
    min_x1 = np.floor(np.min(training[:, x1_idx])) - 1.0
    max_x2 = np.ceil(np.max(training[:, x2_idx])) + 1.0
    min_x2 = np.floor(np.min(training[:, x2_idx])) - 1.0

    xrange = (min_x1, max_x1)
    yrange = (min_x2, max_x2)

    # generate grid coordinates. This will be the basis of the decision
    # boundary visualization.
    (x1, x2) = np.meshgrid(np.arange(xrange[0], xrange[1] + inc / 100, inc),
                           np.arange(yrange[0], yrange[1] + inc / 100, inc))

    # size of the (x1, x2) image, which will also be the size of the
    # decision boundary image that is used as the plot background.
    image_size = x1.shape
    # make (x1, x2) pairs as a bunch of row vectors.
    grid_2d = np.hstack((x1.reshape(x1.shape[0] * x1.shape[1], 1, order='F'),
                         x2.reshape(x2.shape[0] * x2.shape[1], 1, order='F')))

    # Labels for each (x1, x2) pair.
    pred_label = decision_function(grid_2d)

    decision_map = pred_label.reshape(image_size, order='F')

    plt.imshow(decision_map, cmap=plt.cm.Paired,
              extent=[xrange[0], xrange[1], yrange[0], yrange[1]],
              origin='lower')

    # plot the class training data.
    class_names = ['Class ' + str(int(c+1)) for c in range(nclass)]
    data_point_styles = ['rx', 'bo', 'g*']
    color = iter(cm.rainbow(np.linspace(0, 1, nclass)))
    for i in range(nclass):
        plt.scatter(training[label_train == i, x1_idx],
                training[label_train == i, x2_idx],
                color=next(color),
                label=class_names[i])
    plt.legend()

    plt.tight_layout()
    plt.show()


from setuptools import setup
from setuptools import find_packages

with open("README.rst", "r") as f:
    long_description = f.read()

setup(
    name='MLplayground',              # package name
    version='1.0.1',            # package version
    description='A mini Machine Learning Algorithm Library',
    long_description=long_description,
    author='LujiaZhong',
    url='https://lujiazho.github.io/',
    install_requires=[
        "numpy==1.21.1",
        "matplotlib==3.5.1"
    ],
    license='MIT License',
    packages=find_packages(),
    platforms=["Windows 10"],
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
)


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import random
from .Plot import *
import matplotlib.animation as animation

class SMoSVC(Plot):
    def __init__(self, z=[], u=[], b=1, val=[], val_z=[]):
        self.z = np.array(z)          # classes, only for two: 1 or -1
        self.u = np.array(u)          # data
        self.N = len(z)               # number of points

        # validation data
        self.val = np.array(val)
        self.val_z = np.array(val_z)
        
        # parameters for calc lambda
        self.b = b                    # margin
        self.lambda_ = []             # lambda
        
        # parameters for calc w
        self.w = []                   # optimal w
        self.w0 = None                # optimal bias term w0
        
        # kernel types
        self.s_kernels = ['linear']
        self.c_kernels = ['RBF', 'quadratic', 'polynomial', 'sigmoid']

    # n should be not equal to m
    def random_n(self, m):
        n = random.randint(0, self.N-1)
        while n == m:
            n = random.randint(0, self.N-1)
        return n
    
    # for complex kernel
    def K_ij(self, i, j):
        if self.kernel == 'RBF': # RBF kernel's Φ(x) is infinite dimensional
            return np.exp(-self.gamma*np.sum((self.u[i]-self.u[j])*(self.u[i]-self.u[j])))
        if self.kernel == 'quadratic':
            return (self.gamma*np.dot(self.u[i], self.u[j])+self.r)**2
        if self.kernel == 'polynomial':
            return (self.gamma*np.dot(self.u[i], self.u[j])+self.r)**self.d
        if self.kernel == 'sigmoid':
            return np.tanh(self.gamma*np.dot(self.u[i], self.u[j])+self.r)
        return np.dot(self.u[i], self.u[j])
    
    # matrix version of K_ij for accelerating computation, x could be one or two dim
    def K_(self, x):
        if self.kernel == 'RBF':
            return np.exp(-self.gamma*((np.sum(self.u*self.u, axis=1) - 2*np.matmul(self.u, x.T).T).T + np.sum(x*x, axis=len(x.shape)-1)))
        if self.kernel == 'quadratic':
            return (self.gamma*np.matmul(self.u, x.T)+self.r)**2
        if self.kernel == 'polynomial':
            return (self.gamma*np.matmul(self.u, x.T)+self.r)**self.d
        if self.kernel == 'sigmoid':
            return np.tanh(self.gamma*np.matmul(self.u, x.T)+self.r)
        return np.matmul(self.u, x.T)
    
    def f_k(self, k):
        return np.sum((self.lambda_*self.z*self.K_(self.u[k]).T).T, axis=0) + self.w0
    
    def E_k(self, k):
        return self.f_k(k) - self.z[k]*self.b
    
    def crop_lambda_n(self, m, n, lambda_n):
        old_n, old_m = self.lambda_[n], self.lambda_[m]
        if self.z[m] != self.z[n]:
            return min(max(max(0, old_n - old_m), lambda_n), min(self.C, self.C + old_n - old_m))
        else:
            return min(max(max(0, old_m + old_n - self.C), lambda_n), min(old_m + old_n, self.C))
    
    def update_w(self):
        self.w = np.sum(np.multiply(self.lambda_*self.z, self.u.T).T, axis=0)
        
    def update_w0_1(self, E, m, n, o_m, n_m, o_n, n_n):
        return self.w0 - E + self.z[m]*self.K_ij(m,m)*(o_m-n_m) + self.z[n]*self.K_ij(n,m)*(o_n-n_n)
    
    def update_w0_2(self, E, m, n, o_m, n_m, o_n, n_n):
        return self.w0 - E + self.z[m]*self.K_ij(m,n)*(o_m-n_m) + self.z[n]*self.K_ij(n,n)*(o_n-n_n)
    
    # deprecated
    # def k_quadratic(self, x):
    #     return np.array(list(map(lambda x: [x[0],x[1],x[0]*x[0], x[1]*x[1], x[0]*x[1]], x)))
    
    # only for kernel that can be easily stated as K(x,y) = Φ(x)*Φ(y)
    def apply_kernel(self):
        if self.kernel == 'linear':
            return
        # deprecated
        # if self.kernel == 'quadratic':
        #     self.u = self.k_quadratic(self.u)
    
    def fit(self, max_epoch=100, kernel='linear', C=1.0, epsilon=0.0001, ani=False, gamma=3, d=3, r=1):
        """
        Args:
            max_epoch: Maximum training epoch if linear KKT cannot be satisfied.
            kernel: Kernel function. Default is linear.
            C: Upper Limit of lambda. A parameter of soft margin/slack-SVM.
            epsilon: Threshold of lambda value change, if which is less than this & KKT is satisfied, then traning ends in advance.
            ani: Launch traning process record as .gif file. This may cause much slower traning because of drawing in each epoch.
            gamma: Parameter of RBF/quadratic/poly/sigmoid kernels.
            d: Parameter of polynomial kernel.
            r: Parameter of poly/sigmoid kernels.

        Returns:
            Training Parameters & results: [max_epoch, kernel, C, epsilon, ani, gamma, d, acc_train, acc_val]
        """
        self.kernel = kernel
        self.C = C
        self.ani = ani
        self.gamma = gamma
        self.d = d
        self.r = r

        if kernel not in self.s_kernels and kernel not in self.c_kernels:
            print("[Error] No such kernel type!")
            return
        self.apply_kernel()                    # check for simple kernel
        
        self.ims = []                          # for ani

        val_accs = [0]
        max_acc = 0                        # record the maximum and store it's lambda and w0
        self.best_lamd = np.zeros((self.N))
        self.best_w0 = 0
        
        self.lambda_ = np.zeros((self.N))      # initialize lambda
        self.w0 = 0                            # initialize w0
        while max_epoch:
            prev_lambda = np.copy(self.lambda_)
            for m in range(self.N):
                # step 1: randomly choose n != m
                n = self.random_n(m)
                
                old_n, old_m = self.lambda_[n], self.lambda_[m]
                
                # step 2: calculate new lambda_n
                Em, En = self.E_k(m), self.E_k(n)
                eta = self.K_ij(m,m) + self.K_ij(n,n) - 2*self.K_ij(m,n)
                if not eta: # in case there are two points have exactly the same coordinates
                    continue
                lambda_n = old_n + self.z[n]*(Em-En)/eta
                
                # step 3: crop new lambda_n
                lambda_n = self.crop_lambda_n(m, n, lambda_n)
                # if lambda_n < 1e-9, it should be zero because it's not support vector. So we skip it
                # if we don't, the little change of λn can cause big change of result, leading to wrong classification
                if lambda_n < 1e-9:
                    continue
                
                # step 4: calculate new lambda_m
                lambda_m = old_m - self.z[m]*self.z[n]*(lambda_n - old_n)
                
                # step 5: calculate new w0 (slack form)
                if self.C > lambda_m > 0:
                    w0 = self.update_w0_1(Em, m, n, old_m, lambda_m, old_n, lambda_n)
                elif self.C > lambda_n > 0:
                    w0 = self.update_w0_2(En, m, n, old_m, lambda_m, old_n, lambda_n)
                else:
                    cond = (self.lambda_>1e-5)&(self.lambda_<self.C)
                    lam = self.lambda_[cond]
                    if len(lam):
                        w0 = self.b*self.z[cond][0] - np.sum((self.lambda_*self.z*self.K_(self.u[cond][0]).T).T, axis=0)
                    else: # No available point for W0, we adopt no slack case form
                        w0_1 = self.update_w0_1(Em, m, n, old_m, lambda_m, old_n, lambda_n)
                        w0_2 = self.update_w0_2(En, m, n, old_m, lambda_m, old_n, lambda_n)
                        w0 = (w0_1 + w0_2)/2
                
                # update all together
                self.lambda_[m] = lambda_m
                self.lambda_[n] = lambda_n
                self.w0 = w0
            
            # use val data to test model each epoch and store the best lambda and w0
            if self.val.shape[0] and self.val_z.shape[0]:
                cur_res = np.sum((self.lambda_*self.z*self.K_(self.val).T).T, axis=0) + self.w0
                cur_res[cur_res>0] = 1
                cur_res[cur_res<0] = -1

                cur_acc = round(np.sum(cur_res == self.val_z)/len(self.val_z), 5)
                val_accs.append(cur_acc)
                if cur_acc > max_acc:
                    max_acc = cur_acc
                    self.best_lamd = np.copy(self.lambda_)
                    self.best_w0 = self.w0

            # we don't update w, just use kernel for everywhere w is needed
            if self.kernel == 'linear':
                self.update_w() # for plot margin in 2D
    
            if self.ani:
                self.plot()
            
            # check linearly separable case (no slack) KKT
            if np.linalg.norm(prev_lambda - self.lambda_) < epsilon and self.check_KKT_lambda() and self.check_KKT_w_w0():
                print(f"KKT Satisfied!")
                break
            
            max_epoch -= 1
            if max_epoch % 100 == 0 and max_epoch:
                print(f"Last {max_epoch} Epochs")
        if self.ani:
            ani = animation.ArtistAnimation(self.fig, self.ims, interval=200, repeat_delay=1000)
            ani.save(f"animation_{self.kernel}.gif", writer='pillow')
            self.ani = False
        return [max_epoch, kernel, C, epsilon, ani, gamma, d, self.accuracy(), np.max(val_accs)]
    
    # accuracy on traning set
    def accuracy(self):
        result = self.predict(self.u)
        return round(np.sum(result == self.z)/self.N, 5)

    def predict(self, x):
        if self.val.shape[0] and self.val_z.shape[0]:
            result = np.sum((self.best_lamd*self.z*self.K_(x).T).T, axis=0) + self.best_w0
        else:
            result = np.sum((self.lambda_*self.z*self.K_(x).T).T, axis=0) + self.w0
        result[result>0] = 1
        result[result<0] = -1
        return result

    def Error(self, idx, num, track):
        if idx == -1:
            return True
        if not track:
            return False
        if idx == 1:
            print(f"[Check Error] Lambda ({num}) is less than zero!!!")
        elif idx == 2:
            print(f"[Check Error] Sum of product of lambda and z ({num}) is not zero!!!")
        elif idx == 3:
            print(f"[Check Error] λi*[Zi(w*ui + w0)-b] != 0 ({num})!!!")
        elif idx == 4:
            print(f"[Check Error] Zi(w*ui + w0)-b < 0 ({num})!!!")
        elif idx == 5:
            print(f"[Check Error] W* != λi*Zi*ui diff={num}!!!")
        return False

    # only checking linearly separable case (no slack) KKT
    def check_KKT_lambda(self, track=False):
        sum_lambda_z = 0
        for i in range(self.N):
            sum_lambda_z += self.lambda_[i]*self.z[i]
            if round(self.lambda_[i], 9) < 0:
                return self.Error(1, self.lambda_[i], track)
        if abs(sum_lambda_z) > 1e-10:
            return self.Error(2, sum_lambda_z, track)
        return self.Error(-1, 0, track)

    # only checking linearly separable case (no slack) KKT
    def check_KKT_w_w0(self, track=False):
#         sum_lambda_z_u = np.full(self.u.shape[1], 0.0)
        for i in range(self.N):
#             check1 = self.lambda_[i]*(self.z[i]*(np.matmul(self.w, self.u[i]) + self.w0)-self.b)
            check1 = self.lambda_[i]*(self.z[i]*self.f_k(i)-self.b)
            if abs(check1) > 1e-10:
                return self.Error(3, check1, track)
            check2 = self.z[i]*self.f_k(i) - self.b
            check2 = 0 if abs(check2) < 1e-10 else check2
            if check2 < 0:
                return self.Error(4, check2, track)
#             sum_lambda_z_u += self.lambda_[i]*self.z[i]*self.u[i]
#         if sum(sum_lambda_z_u != self.w):
#             return self.Error(5, sum_lambda_z_u-self.w, track)
        return self.Error(-1, 0, track)


#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import numpy as np

# primary node
class Node:
    def __init__(self, inputs=[]):
        self.inputs = inputs
        self.outputs = []

        for n in self.inputs:
            n.outputs.append(self)

        self.value = None
        self.gradients = {}

    def forward(self):
        raise NotImplemented
    
    def backward(self):
        raise NotImplemented

    def optimize(self):
        raise NotImplemented

# parameter node
class Placeholder(Node):
    def __init__(self):
        Node.__init__(self)
         # For Adam optimizer
        self.mt = 0
        self.vt = 0

    def forward(self, value=None):
        if value is not None:
            self.value = value
        
    def backward(self):
        self.gradients = {self:0}

        for n in self.outputs:
            grad_cost = n.gradients[self]
            self.gradients[self] += grad_cost * 1

    def optimize(self, optimizer='SGD', lr=0.001, it=0):
        beta1, beta2 = 0.9, 0.999
        epsilon = 1e-8
        if optimizer == 'SGD':
            self.value -= lr * self.gradients[self]
        elif optimizer == 'Adam':
            self.mt = beta1*self.mt + (1-beta1)*self.gradients[self]
            mt_hat = self.mt/(1-beta1**it)
            self.vt = beta2*self.vt + (1-beta2)*self.gradients[self]**2
            vt_hat = self.vt/(1-beta2**it)
            self.value -= lr * mt_hat / (epsilon + vt_hat**0.5)
        elif optimizer == 'RAdam':
            ro_inf = 2/(1-beta2) - 1
            ro_t = ro_inf - 2*it*beta2**it/(1-beta2**it)

            self.mt = beta1*self.mt + (1-beta1)*self.gradients[self]
            mt_hat = self.mt/(1-beta1**it)
            if ro_t <= 4:
                self.value -= lr * mt_hat
            else:
                r_t = np.sqrt((ro_t-4)*(ro_t-2)*ro_inf/((ro_inf-4)*(ro_inf-2)*ro_t))                            
                self.vt = beta2*self.vt + (1-beta2)*self.gradients[self]**2
                vt_hat = self.vt/(1-beta2**it)

                self.value -= lr * mt_hat * r_t / (epsilon + vt_hat**0.5)

# functional node: add all inputs' values
class Add(Node):
    def __init__(self, *nodes):
        Node.__init__(self, list(nodes))

    def forward(self):
        self.value = sum(map(lambda n: n.value, self.inputs))

    def backward(self):
        # create zero loss for all inputs
        self.gradients = {n: np.zeros_like(n.value) for n in self.inputs}

        for n in self.outputs:
            # Get the cost w.r.t this node.
            grad_cost = n.gradients[self]

            for i in range(len(self.inputs)-1):
                self.gradients[self.inputs[i]] += grad_cost
            self.gradients[self.inputs[-1]] += np.sum(grad_cost, axis=0, keepdims=False)

# functional node: calculate linear function
class Linear(Node):
    def __init__(self, nodes, weights, bias):
        self.bias = True if bias else False
        Node.__init__(self, [nodes, weights, bias] if bias else [nodes, weights])

    def forward(self):
        inputs = self.inputs[0].value
        weights = self.inputs[1].value
        if self.bias:
            bias = self.inputs[2].value

        self.value = np.dot(inputs, weights) + (bias if self.bias else 0)
        
    def backward(self):
        # create zero loss for all inputs
        self.gradients = {n: np.zeros_like(n.value) for n in self.inputs}

        for n in self.outputs:
            # Get the cost w.r.t this node.
            grad_cost = n.gradients[self]

            self.gradients[self.inputs[0]] += np.dot(grad_cost, self.inputs[1].value.T)
            self.gradients[self.inputs[1]] += np.dot(self.inputs[0].value.T, grad_cost)
            if self.bias:
                self.gradients[self.inputs[2]] += np.sum(grad_cost, axis=0, keepdims=False)

# simple embedding for nlp
class Embedding(Node):
    def __init__(self, node, embedding):
        Node.__init__(self, [node, embedding])

    def forward(self):
        x = self.inputs[0].value
        embedding = self.inputs[1].value

        self.value = embedding[x].reshape(len(x), -1)
        
    def backward(self):
        self.gradients = {n: np.zeros_like(n.value, dtype=np.float64) for n in self.inputs}
        x = self.inputs[0].value # n*c
        embedding = self.inputs[1].value # w*d

        x_ = np.zeros((x.size, embedding.shape[0]))
        x_[np.arange(x.size), x.flatten()] = 1

        for n in self.outputs:
            grad_cost = n.gradients[self]

            grad_R = grad_cost.reshape((x.shape[0]*x.shape[1], embedding.shape[1]))
            self.gradients[self.inputs[1]] += np.dot(x_.T, grad_R)

            grad_x_ = np.dot(grad_R, embedding.T)
            self.gradients[self.inputs[0]] += np.sum(grad_x_, axis=1, keepdims=False).reshape((x.shape[0], x.shape[1]))

# functional node: calculate sigmoid function
class Sigmoid(Node):
    def __init__(self, node):
        Node.__init__(self, [node])

    def _sigmoid(self, x):
        return 1./(1 + np.exp(-1 * x))

    def forward(self):
        self.x = self.inputs[0].value
        self.value = self._sigmoid(self.x)

    def backward(self):
        self.partial = self._sigmoid(self.x) * (1 - self._sigmoid(self.x))
        self.gradients = {n: np.zeros_like(n.value) for n in self.inputs}

        for n in self.outputs:
            grad_cost = n.gradients[self]

            self.gradients[self.inputs[0]] += grad_cost * self.partial

# functional node: calculate relu function
class Relu(Node):
    def __init__(self, node):
        Node.__init__(self, [node])

    def _relu(self, x):
        x[x<0] = 0
        return x

    def _partial(self, x):
        x[x<0] = 0
        x[x>0] = 1
        return x

    def forward(self):
        self.x = self.inputs[0].value
        self.value = self._relu(self.x)

    def backward(self):
        self.gradients = {n: np.zeros_like(n.value) for n in self.inputs}

        for n in self.outputs:
            grad_cost = n.gradients[self]

            self.gradients[self.inputs[0]] += grad_cost * self._partial(self.x)

# functional node: tanh function
class Tanh(Node):
    def __init__(self, node):
        Node.__init__(self, [node])

    def _tanh(self, x):
        return (np.exp(x) - np.exp(-x))/(np.exp(x) + np.exp(-x))

    def forward(self):
        self.x = self.inputs[0].value
        self.value = self._tanh(self.x)

    def backward(self):
        self.partial = (1 - self._tanh(self.x)**2)
        self.gradients = {n: np.zeros_like(n.value) for n in self.inputs}

        for n in self.outputs:
            grad_cost = n.gradients[self]

            self.gradients[self.inputs[0]] += grad_cost * self.partial

# functional node: calculate MSE function
class MSE(Node):
    def __init__(self, y, a):
        Node.__init__(self, [y, a])

    def forward(self):
        y = self.inputs[0].value
        a = self.inputs[1].value
        assert(y.shape == a.shape)

        self.m = self.inputs[0].value.shape[0]
        self.diff = y - a

        self.value = np.mean(self.diff**2)

    # Derivative * loss = Δw = gradients
    def backward(self):
        self.gradients[self.inputs[0]] = (2 / self.m) * self.diff
        self.gradients[self.inputs[1]] = (-2 / self.m) * self.diff



#########################################
#         MLPlayground for Numpy        #
#      Machine Learning Techniques      #
#                 v1.0.0                #
#                                       #
#         Written by Lujia Zhong        #
#       https://lujiazho.github.io/     #
#              26 May 2022              #
#              MIT License              #
#########################################
import random
import numpy as np

class KMeans:
    def __init__(self, data):
        self.data = np.array(data)
    
    def self_cdist(self):
        dis = []
        for i in range(self.k):
            class_ = (np.hstack([self.data for _ in range(self.k)]) - self.means.reshape(1, -1))[:,i*self.dim:(i+1)*self.dim]
            dis.append(np.linalg.norm(class_, axis=1).reshape(-1,1))
        return np.hstack(dis)
    
    def fit(self, k=3):
        if not len(self.data):
            return
        self.dim = len(self.data[0])  # dimension of data points
        self.k = k
        self.means = []
        
        # step 1: initialize k cluster centers with the same dimension as each point
        for _ in range(self.k):
            self.means.append(random.choice(self.data))
        self.means = np.array(self.means)

        iteration = 100
        while iteration:
            # step 2: find the nearest cluster center for each point
            dist_mat = self.self_cdist()
            new_class = np.argmin(dist_mat, axis=1)
            
            # step 3: recalculate cluster center for each cluster corresponding new points assignment
            new_means = []
            for i in range(self.k):
                new_means.append(np.mean(self.data[new_class==i], axis=0))
            new_means = np.array(new_means)
            
            # step 4: if means didn't change, algorithm stops
            if np.sum(new_means == self.means) == self.dim*self.k:
                break
            self.means = new_means
            iteration -= 1
        return [iteration, self.k]

    def predict(self, inputs):
        pred = []
        for i, xy in enumerate(inputs):
            my_eu = []
            for j in range(self.c):
                # or np.linalg.norm(xy - self.sample_means[j])
                my_eu.append(np.sqrt(np.sum(np.square(xy - self.means[j]))))
            pred.append(np.argmin(my_eu))
        return np.array(pred)